
#ifdef WIN32
#include <windows.h>
#include <process.h>
#else
#include <pthread.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sql.h>        // core
#include <sqlext.h>     // extensions
#include <string>

using namespace std;

// Thread mutex
#ifdef WIN32
	HANDLE m_mMutex;
#else
	pthread_mutex_t m_mMutex;
#endif

// End Thread flag
int EndThread = 0;
int NumErrors = 0;

// Num threads
int NumThreads = 1;

SQLCHAR	Source[255];	// Data source name
SQLCHAR	User[255];		// Data source user
SQLCHAR	Password[255];	// Data source password
SQLCHAR	Table[255];		// Data source table

SQLHENV     henv;	/* environment handle */
SQLHDBC     hdbc;	/* connection handle */
SQLHSTMT    hstmt;	/* statement handle */


void DisplaySqlError(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt);
int DisplaySqlStringInfo(SQLHENV henv, SQLHDBC hdbc, SQLUSMALLINT InfoType);
int DoConnect();
int DoDisconnect();
int GetCurrentThreadId();

#ifdef WIN32
void __cdecl ThreadMain(void *Value);
#else
void* ThreadMain(void *Value);
#endif


int main(int argc, char* argv[])
{
	if(argc <=1)
	{
		printf("usage: testodbc2 -s DataSource -u User [-p Password]\n");
		return 0;
	}
	Source[0] = '\0';
	User[0] = '\0';
	Password[0] = '\0';
	int i;
	for(i=1;i<argc;i++)
	{
		if(strcmp(argv[i], "-s") == 0)
		{
			if(i==(argc-1))
			{
				printf("usage: testodbc2 -s DataSource -u User [-p Password]\n");
				return 0;
			}
			i++;
			strcpy((char *) Source, argv[i]);
		}
		else if(strcmp(argv[i], "-u") == 0)
		{
			if(i==(argc-1))
			{
				printf("usage: testodbc2 -s DataSource -u User [-p Password]\n");
				return 0;
			}
			i++;
			strcpy((char *) User, argv[i]);
		}
		else if(strcmp(argv[i], "-p") == 0)
		{
			if(i==(argc-1))
			{
				printf("usage: testodbc2 -s DataSource -u User [-p Password]\n");
				return 0;
			}
			i++;
			strcpy((char *) Password, argv[i]);
		}
	}

	if(!DoConnect())
		return 0;

#ifdef WIN32
	m_mMutex = CreateMutex(NULL, 0, NULL);
#else
	pthread_mutex_init(&m_mMutex, NULL);
#endif


	EndThread = NumThreads;
	int EndThreadHere = EndThread;
	for(i=0;i<NumThreads;i++)
	{
#ifdef WIN32
		_beginthread(ThreadMain, 0, NULL);
#else
		pthread_attr_t	attr;
		pthread_attr_init(&attr);
		pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);		// the unix thread has a system scope for best performances
		pthread_t thid;
		pthread_create(&thid, &attr, ThreadMain, NULL);
		pthread_attr_destroy(&attr);
#endif
	}
	while(EndThreadHere)
	{
#ifdef WIN32
		WaitForSingleObject(m_mMutex, INFINITE);
#else
		pthread_mutex_lock(&m_mMutex);
#endif
		EndThreadHere = EndThread;
#ifdef WIN32
		ReleaseMutex(m_mMutex);
#else
		pthread_mutex_unlock(&m_mMutex);
#endif
	}

	if(NumErrors)
		printf("\nProgram terminated with %d errors\n", NumErrors);
	else
		printf("\nProgram terminated successfully\n");

#ifdef WIN32
	ReleaseMutex(m_mMutex);
	CloseHandle(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
	pthread_mutex_destroy(&m_mMutex);
#endif

	return 0;
}


#ifdef WIN32
void __cdecl ThreadMain(void *Value)
#else
void* ThreadMain(void *Value)
#endif
{
	int RetVal = DoDisconnect();

#ifdef WIN32
	WaitForSingleObject(m_mMutex, INFINITE);
#else
	pthread_mutex_lock(&m_mMutex);
#endif
	if(!RetVal)
		NumErrors++;
	EndThread--;
#ifdef WIN32
	ReleaseMutex(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
#endif

#ifdef WIN32
	return;						// end of thread (windows)
#else
	pthread_detach(pthread_self());
	pthread_exit(NULL);

	return (void *)(0);			// end of thread (unix)
#endif
}

int DoConnect()
{
	SQLRETURN   retcode;	/* return code */
	SQLCHAR SQLStmt[255];

	henv = NULL;
	hdbc = NULL;
	hstmt = NULL;

	printf("Open db source %s from thread %d\n", Source, GetCurrentThreadId()
);

	retcode = SQLAllocEnv(&henv);
	retcode = SQLAllocConnect(henv, &hdbc);
	retcode = SQLConnect(hdbc, Source, SQL_NTS, User, SQL_NTS, Password, SQL_NTS);

	if (retcode != SQL_SUCCESS)
	{
		printf("*** Cannot connect data source\n");
		DisplaySqlError(henv, hdbc, hstmt);
		SQLFreeConnect(hdbc);
		SQLFreeEnv(henv);
		return 0;
	}
	printf("Source successfully connected\n");
	printf("Database name: ");
	DisplaySqlStringInfo(henv, hdbc,SQL_DBMS_NAME);
	printf("\nDatabase version: ");
	DisplaySqlStringInfo(henv, hdbc,SQL_DBMS_VER);
	printf("\nDriver name: ");
	DisplaySqlStringInfo(henv, hdbc,SQL_DRIVER_NAME);
	printf("\nDriver version: ");
	DisplaySqlStringInfo(henv, hdbc,SQL_DRIVER_VER);
	printf("\nDriver ODBC version: ");
	DisplaySqlStringInfo(henv, hdbc,SQL_ODBC_VER);
	printf("\n");

	return 1;
}



int DoDisconnect()
{

	printf("Close db source from thread %d\n", GetCurrentThreadId());
	SQLDisconnect(hdbc);
	SQLFreeConnect(hdbc);
	SQLFreeEnv(henv);
	printf("Source successfully disconnected\n");

	return 1;
}


void DisplaySqlError(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
	SQLCHAR	Message[SQL_MAX_MESSAGE_LENGTH + 1];
	SQLCHAR	State[10];
	SQLINTEGER NativeError;
	SQLSMALLINT TextLength;

	Message[0] = '\0';
	SQLError(henv, hdbc, hstmt, State, &NativeError, Message, SQL_MAX_MESSAGE_LENGTH+1, &TextLength);

	printf("SQL error: %s\n", (const char *)Message);
}


int DisplaySqlStringInfo(SQLHENV henv, SQLHDBC hdbc, SQLUSMALLINT InfoType)
{
	SQLRETURN   retcode;	/* return code */
	SQLCHAR    info[255];	/* info string for SQLGetInfo */
	SQLSMALLINT cbInfoValue;

	info[0] = '\0';
	retcode = SQLGetInfo(hdbc,InfoType, &info, sizeof(info),&cbInfoValue);
	if (retcode == SQL_SUCCESS)
		printf((const char *)info);
	else
	{
		printf("*** Cannot get info from data source\n");
		DisplaySqlError(henv, hdbc, NULL);
		return 0;
	}
	return 1;
}

int GetCurrentThreadId()

{

	int CurrentThreadId = 0;

#ifdef WIN32

	CurrentThreadId = (int)GetCurrentThreadId();			// Thread handle for window

#else

	CurrentThreadId = (int)pthread_self();				// Thread handle for unix

#endif

	return CurrentThreadId;

}